# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import os
import sys

import pip
import pkg_resources
from flask import current_app, render_template
from pathlib import Path
from pip.vcs import vcs

from ..panel import AdminPanel


class SysinfoPanel(AdminPanel):
    id = 'sysinfo'
    label = 'System information'
    icon = 'hdd'

    def get(self):
        uname = os.popen("uname -a").read()
        python_version = sys.version.strip()

        packages = []

        for dist in pkg_resources.working_set:
            package = dict(name=dist.project_name,
                           key=dist.key,
                           version=dist.version if dist.has_version(
                           ) else u'Unknown version',
                           vcs=None,)

            location = unicode(Path(dist.location).absolute())
            vcs_name = vcs.get_backend_name(location)

            if vcs_name:
                vc = vcs.get_backend(vcs_name)()
                try:
                    url = vc.get_url(location)
                except pip.exceptions.InstallationError:
                    url = 'None'
                try:
                    revision = vc.get_revision(location)
                except pip.exceptions.InstallationError:
                    revision = 'None'
                package['vcs'] = dict(name=vcs_name, url=url, revision=revision)

            packages.append(package)
            packages.sort(key=lambda d: d.get('key'))

        config_values = [(k, repr(v))
                         for k, v in sorted(current_app.config.items())]

        return render_template("admin/sysinfo.html",
                               python_version=python_version,
                               packages=packages,
                               uname=uname,
                               config_values=config_values)



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import logging
from functools import wraps

from flask import current_app

from abilian.core.util import fqcn


class ServiceNotRegistered(Exception):
    pass


class ServiceState(object):
    """
    Service state stored in Application.extensions
    """
    #: reference to :class:`Service` instance
    service = None

    running = False

    def __init__(self, service, running=False):
        self.service = service
        self.running = running
        self.logger = logging.getLogger(fqcn(self.__class__))


class Service(object):
    """
    Base class for services.
    """
    #: State class to use for this Service
    AppStateClass = ServiceState

    #: service name in Application.extensions / Application.services
    name = None

    def __init__(self, app=None):
        if self.name is None:
            raise ValueError('Service must have a name ({})'.format(fqcn(
                self.__class__)))

        self.logger = logging.getLogger(fqcn(self.__class__))
        if app:
            self.init_app(app)

    def init_app(self, app):
        app.extensions[self.name] = self.AppStateClass(self)
        app.services[self.name] = self

    def start(self, ignore_state=False):
        """
        Starts the service.
        """
        self.logger.debug('Start service')
        self._toggle_running(True, ignore_state)

    def stop(self, ignore_state=False):
        """
        Stops the service.
        """
        self.logger.debug('Stop service')
        self._toggle_running(False, ignore_state)

    def _toggle_running(self, run_state, ignore_state=False):
        state = self.app_state
        run_state = bool(run_state)
        if not ignore_state:
            assert run_state ^ state.running
        state.running = run_state

    @property
    def app_state(self):
        """ Current service state in current application.

        :raise:RuntimeError if working outside application context.
        """
        try:
            return current_app.extensions[self.name]
        except KeyError:
            raise ServiceNotRegistered(self.name)

    @property
    def running(self):
        """
        :returns: `False` if working outside application context, if service is
                  not registered on current application,  or if service is halted
                  for current application.
        """
        try:
            return self.app_state.running
        except (RuntimeError, ServiceNotRegistered):
            # RuntimeError: happens when current_app is None: working outside
            # application context
            return False

    @staticmethod
    def if_running(meth):
        """
        Decorator for service methods that must be ran only if service is in
        running state.
        """

        @wraps(meth)
        def check_running(self, *args, **kwargs):
            if not self.running:
                return
            return meth(self, *args, **kwargs)

        return check_running



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

from cgi import escape

import sqlalchemy as sa
from flask import current_app, render_template_string, request
from flask_babel import format_datetime
from flask_login import current_user
from sqlalchemy.sql.expression import asc, desc, func, nullslast, or_
from werkzeug.datastructures import MultiDict

from abilian.core.models.subjects import User, gen_random_password
from abilian.i18n import _
from abilian.services.security.models import Admin, Role
from abilian.web.nav import BreadcrumbItem
from abilian.web.util import url_for
from abilian.web.views import object as views
from abilian.web.views import base
from abilian.web.views.images import user_photo_url

from .forms import UserAdminForm, UserCreateForm


class JsonUsersList(base.JSONView):
    """JSON user list for datatable.
    """

    def data(self, *args, **kw):
        security = current_app.services['security']
        length = int(kw.get("iDisplayLength", 0))
        start = int(kw.get("iDisplayStart", 0))
        sort_col = int(kw.get("iSortCol_0", 1))
        sort_dir = kw.get("sSortDir_0", "asc")
        echo = int(kw.get("sEcho", 0))
        search = kw.get("sSearch", "").replace("%", "").strip().lower()

        end = start + length
        q = User.query \
          .options(sa.orm.subqueryload('groups'),
                   sa.orm.undefer('photo'), ) \
          .filter(User.id != 0)
        total_count = q.count()

        if search:
            # TODO: gérer les accents
            filter = or_(
                func.lower(User.first_name).like("%" + search + "%"),
                func.lower(User.last_name).like("%" + search + "%"),
                func.lower(User.email).like("%" + search + "%"))
            q = q.filter(filter)

        count = q.count()
        SORT_COLS = {
            1: [],  # [User.last_name, User.first_name] will be added anyway
            2: [func.lower(User.email)],
            5: [User.last_active],
        }
        columns = list(SORT_COLS.get(sort_col, []))
        columns.extend([func.lower(User.last_name), func.lower(User.first_name)
                       ])

        direction = asc if sort_dir == 'asc' else desc
        order_by = map(direction, columns)

        # sqlite does not support 'NULLS FIRST|LAST' in ORDER BY clauses
        engine = q.session.get_bind(User.__mapper__)
        if engine.name != 'sqlite':
            order_by[0] = nullslast(order_by[0])

        q = q.order_by(*order_by)

        users = q.slice(start, end).all()

        data = []
        MUGSHOT_SIZE = 45
        for user in users:
            # TODO: this should be done on the browser.
            user_url = url_for(".users_user", user_id=user.id)
            mugshot = user_photo_url(user, size=MUGSHOT_SIZE)
            name = escape(getattr(user, "name") or "")
            email = escape(getattr(user, "email") or "")
            roles = [r
                     for r in security.get_roles(user, no_group_roles=True)
                     if r.assignable]
            columns = []
            columns.append(
                u'<a href="{url}"><img src="{src}" width="{size}" height="{size}">'
                u'</a>'.format(url=user_url,
                               src=mugshot,
                               size=MUGSHOT_SIZE))
            columns.append(u'<a href="{url}">{name}</a>'.format(url=user_url,
                                                                name=name))
            columns.append(u'<a href="{url}"><em>{email}</em></a>'.format(
                url=user_url, email=email))
            columns.append(u'\u2713' if user.can_login else u'')
            columns.append(render_template_string(u'''{%- for g in groups %}
            <span class="badge badge-default">{{ g.name }}</span>
            {%- endfor %}''',
                                                  groups=sorted(user.groups)))
            columns.append(render_template_string(u'''{%- for role in roles %}
            <span class="badge badge-default">{{ role }}</span>
            {%- endfor %}''',
                                                  roles=roles))

            if user.last_active:
                last_active = format_datetime(user.last_active)
            else:
                last_active = _(u'Never logged in')
            columns.append(last_active)

            data.append(columns)

        return {
            "sEcho": echo,
            "iTotalRecords": total_count,
            "iTotalDisplayRecords": count,
            "aaData": data,
        }


# User edit / create views
class UserBase(object):
    Model = User
    pk = 'user_id'
    Form = UserAdminForm
    base_template = 'admin/_base.html'

    def index_url(self):
        return url_for('.users')

    view_url = index_url


class UserEdit(UserBase, views.ObjectEdit):

    def breadcrumb(self):
        label = render_template_string(u'<em>{{ u.email }}</em>', u=self.obj)
        return BreadcrumbItem(label=label, url=u'', description=self.obj.name)

    def get_form_kwargs(self):
        kw = super(UserEdit, self).get_form_kwargs()
        security = current_app.services['security']
        roles = security.get_roles(self.obj, no_group_roles=True)
        kw['roles'] = [r.name for r in roles if r.assignable]
        return kw

    def validate(self):
        if not super(UserEdit, self).validate():
            return False

        if current_user == self.obj:
            # self edit: don't let user shoot themself in the foot
            self.form.can_login.data = True
            roles = self.form.roles
            if Admin.name not in roles.data:
                roles.data.append(Admin.name)
        return True

    def form_valid(self):
        del self.form.confirm_password
        if self.form.password.data:
            self.obj.set_password(self.form.password.data)
        del self.form.password

        return super(UserEdit, self).form_valid()

    def after_populate_obj(self):
        security = current_app.services['security']
        current_roles = security.get_roles(self.obj, no_group_roles=True)
        current_roles = set(r for r in current_roles if r.assignable)
        new_roles = {Role(r) for r in self.form.roles.data}

        for r in (current_roles - new_roles):
            security.ungrant_role(self.obj, r)

        for r in (new_roles - current_roles):
            security.grant_role(self.obj, r)

        return super(UserEdit, self).after_populate_obj()


class UserCreate(UserBase, views.ObjectCreate):
    Form = UserCreateForm
    chain_create_allowed = True

    def get_form_kwargs(self):
        kw = super(UserCreate, self).get_form_kwargs()

        if request.method == 'GET':
            # ensure formdata is not ImmutableMultiDict (request.args)
            data = MultiDict(kw.setdefault('formdata', {}))
            kw['formdata'] = data
            data['can_login'] = True

        return kw

    def form_valid(self):
        if not self.form.password.data:
            self.form.password.data = gen_random_password()

        self.obj.set_password(self.form.password.data)
        del self.form.password
        return super(UserCreate, self).form_valid()



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import logging
from collections import OrderedDict
from functools import partial

import whoosh
import whoosh.sorting
from flask import Blueprint, current_app, g, render_template, request, url_for

from abilian.i18n import _
from abilian.web import views
from abilian.web.action import Endpoint
from abilian.web.nav import BreadcrumbItem

logger = logging.getLogger(__name__)

BOOTSTRAP_MARKUP_HIGHLIGHTER = whoosh.highlight.HtmlFormatter(
    tagname=u'mark',
    classname=u'',
    termclass=u'term-',
    between=u'[…]')

RESULTS_FRAGMENTER = whoosh.highlight.SentenceFragmenter()

#
# Some hardcoded settings (for now)
#
PAGE_SIZE = 20
MAX_LIVE_RESULTS_PER_CLASS = 5

search = Blueprint('search',
                   __name__,
                   url_prefix="/search",
                   template_folder='templates')
route = search.route


def friendly_fqcn(fqcn):
    return fqcn.rsplit(u'.', 1)[-1]


@search.url_value_preprocessor
def init_search(endpoint, values):
    q = request.args.get("q")
    try:
        page = max(int(request.args.get('page', 1), 1))
    except:
        page = 1

    g.breadcrumb.append(BreadcrumbItem(
        label=u'"{}"'.format(q),
        icon="search",
        url=Endpoint('search.search_main', q=q)))

    page_kw = OrderedDict(q=q)
    object_types = request.args.getlist('object_type')

    if object_types:
        page_kw['object_type'] = object_types
        g.breadcrumb.append(BreadcrumbItem(
            label=u' | '.join(friendly_fqcn(name) for name in object_types),
            url=Endpoint('search.search_main', **page_kw)))

    if page > 1:
        g.breadcrumb.append(
            BreadcrumbItem(label=unicode(page),
                               url=Endpoint('search.search_main',
                                                page=page,
                                                **page_kw)))

    values['q'] = q
    values['page'] = page


@search.context_processor
def install_hit_to_url():
    return dict(url_for_hit=current_app.extensions['indexing'].url_for_hit)


_COUNT_OBJECT_TYPE_FACET = whoosh.sorting.FieldFacet(
    "object_type",
    maptype=whoosh.sorting.Count,)


@route('')
def search_main(q=u'', page=1):
    svc = current_app.services['indexing']
    q = q.strip()
    page = int(request.args.get('page', page))
    search_kwargs = {'limit': page * PAGE_SIZE}
    page_url_kw = OrderedDict(q=q)

    search_kwargs['groupedby'] = _COUNT_OBJECT_TYPE_FACET
    results = svc.search(q, **search_kwargs)

    filtered_by_type = sorted(request.args.getlist('object_type'))
    if filtered_by_type:
        page_url_kw['object_type'] = filtered_by_type

    page_url = partial(url_for, '.search_main', **page_url_kw)

    # get facets groups
    by_object_type = []
    try:
        facet_group = results.groups('object_type')
    except KeyError:
        pass
    else:
        for typename, count in facet_group.items():
            is_active = typename in filtered_by_type
            classname = friendly_fqcn(typename)
            link = page_url(object_type=typename)
            by_object_type.append((classname, count, link, is_active))

        by_object_type.sort(key=lambda t: t[0])

    if by_object_type:
        # Insert 'all' to clear all filters
        is_active = len(filtered_by_type) == 0
        all_types = (_(u'All'), len(results), page_url(object_type=(),),
                     is_active,)
        by_object_type.insert(0, all_types)

    if filtered_by_type:
        #FIXME: sanitize input
        results = svc.search(q, object_types=filtered_by_type, **search_kwargs)

    results.formatter = BOOTSTRAP_MARKUP_HIGHLIGHTER
    results.fragmenter = RESULTS_FRAGMENTER

    # paginate results
    results_count = len(results)

    # results.pagecount must be ignored when query is filtered: it ignores
    # filtered_count
    pagecount = 1 + results_count // PAGE_SIZE
    page = min(page, pagecount)
    results = whoosh.searching.ResultsPage(results, page, PAGE_SIZE)
    page = results.pagenum
    first_page = page_url(page=1)
    last_page = page_url(page=pagecount)
    prev_page = page_url(page=page - 1) if page > 1 else None
    next_page = page_url(page=page + 1) if page < pagecount else None

    page_min = max(page - 2, 1)
    page_max = min(page + 4, pagecount)
    next_pages_numbered = [(index, page_url(page=index))
                           for index in range(page_min, page_max + 1)]

    return render_template('search/search.html',
                           q=q,
                           results=results,
                           results_count=results_count,
                           pagecount=pagecount,
                           filtered_by_type=filtered_by_type,
                           by_object_type=by_object_type,
                           prev_page=prev_page,
                           next_page=next_page,
                           first_page=first_page,
                           last_page=last_page,
                           next_pages_numbered=next_pages_numbered,
                           friendly_fqcn=friendly_fqcn,)


class Live(views.JSONView):
    """
    JSON response for live search
    """

    def data(self, q=u'', page=None, *args, **kwargs):
        svc = current_app.services['indexing']
        url_for_hit = svc.app_state.url_for_hit
        search_kwargs = {'facet_by_type': 5}
        response = {}
        results = svc.search(q, **search_kwargs)
        datasets = {}

        for typename, docs in results.iteritems():
            dataset = []
            for doc in docs:
                d = dict(name=doc['name'])
                url = url_for_hit(doc, None)
                if url is not None:
                    d['url'] = url
                    dataset.append(d)
            datasets[typename] = dataset

        response['results'] = datasets
        return response


route('/live')(Live.as_view('live'))



from __future__ import absolute_import, division, print_function

import uuid

from sqlalchemy import Column

from abilian.core.entities import Entity
from abilian.core.extensions import db
from abilian.core.sqlalchemy import UUID, JSONDict, JSONList
from abilian.testing import BaseTestCase


class DummyModel2(Entity):
    list_attr = Column(JSONList())
    dict_attr = Column(JSONDict())
    uuid = Column(UUID())


class SAExtensionTestCase(BaseTestCase):

    def test_list_attribute(self):
        model = DummyModel2(list_attr=[1, 2, 3])
        db.session.add(model)
        db.session.commit()
        model_id = model.id
        db.session.remove()
        model2 = DummyModel2.query.get(model_id)
        assert model2.list_attr == [1, 2, 3]

        model2.list_attr.append(4)
        assert model2.list_attr == [1, 2, 3, 4]

        del model2.list_attr[0]
        assert model2.list_attr == [2, 3, 4]

    def test_dict_attribute(self):
        model = DummyModel2(dict_attr=dict(a=3, b=4))
        db.session.add(model)
        db.session.commit()
        model_id = model.id
        db.session.remove()
        model2 = DummyModel2.query.get(model_id)
        assert model2.dict_attr == dict(a=3, b=4)

        model2.dict_attr['c'] = 5
        assert model2.dict_attr == dict(a=3, b=4, c=5)

    def test_uuid_attribute(self):
        # uuid from string
        model = DummyModel2(uuid='c5ad316a-2cd0-4f78-a49b-cff216c10713')
        db.session.add(model)
        db.session.commit()
        model_id = model.id
        db.session.remove()
        model2 = DummyModel2.query.get(model_id)

        assert isinstance(model2.uuid, uuid.UUID)

        # plain UUID object
        u = uuid.UUID('3eb7f164-bf15-4564-a058-31bdea0196e6')
        model = DummyModel2(uuid=u)
        db.session.add(model)
        db.session.commit()
        model_id = model.id
        db.session.remove()
        model2 = DummyModel2.query.get(model_id)

        assert isinstance(model2.uuid, uuid.UUID)
        assert model2.uuid == u



from __future__ import absolute_import, division, print_function

from sqlalchemy import Column, Text, UnicodeText

from abilian.core.entities import SEARCHABLE, Entity


class DummyContact(Entity):
    salutation = Column(UnicodeText, default=u"")
    first_name = Column(UnicodeText, default=u"", info=SEARCHABLE)
    last_name = Column(UnicodeText, default=u"", info=SEARCHABLE)
    email = Column(Text, default=u"")



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import abc
from functools import total_ordering

import sqlalchemy as sa

from abilian.core.entities import Entity

from .base import IdMixin, Model

#: backref attribute on tagged elements
TAGS_ATTR = '__tags__'


class SupportTagging(object):
    __metaclass__ = abc.ABCMeta


def register(cls):
    """Register an :class:`~.Entity` as a taggable class.

    Can be used as a class decorator:

    .. code-block:: python

      @tag.register
      class MyContent(Entity):
          ....
    """
    if not issubclass(cls, Entity):
        raise ValueError(
            'Class must be a subclass of abilian.core.entities.Entity')

    SupportTagging.register(cls)
    return cls


def is_support_tagging(obj):
    """
    :param obj: a class or instance
    """
    if isinstance(obj, type):
        return issubclass(obj, SupportTagging)

    if not isinstance(obj, SupportTagging):
        return False

    if obj.id is None:
        return False

    return True


entity_tag_tbl = sa.Table('entity_tags',
                          Model.metadata,
                          sa.Column('tag_id',
                                    sa.Integer,
                                    sa.ForeignKey('tag.id',
                                                  ondelete='CASCADE')),
                          sa.Column('entity_id',
                                    sa.Integer,
                                    sa.ForeignKey(Entity.id,
                                                  ondelete='CASCADE')),
                          sa.UniqueConstraint('tag_id', 'entity_id'),)


@total_ordering
class Tag(IdMixin, Model):
    """Tags are text labels that can be attached to :class:`entities <.Entity>`.

    They are namespaced, so that independent group of tags can be defined in the
    application. The default namespace is `"default"`.
    """
    __tablename__ = 'tag'

    #: namespace
    ns = sa.Column(sa.UnicodeText(),
                   nullable=False,
                   default=u'default',
                   server_default='default')

    #: Label visible to the user
    label = sa.Column(sa.UnicodeText(), nullable=False)

    #: :class:`entities <.Entity>` attached to this tag
    entities = sa.orm.relationship(
        Entity,
        collection_class=set,
        secondary=entity_tag_tbl,
        backref=sa.orm.backref(TAGS_ATTR, collection_class=set),)

    __mapper_args__ = {'order_by': label,}

    __table_args__ = (
        sa.UniqueConstraint(ns, label),
        # namespace is not empty and is not surrounded by space characters
        sa.CheckConstraint(sa.sql.and_(
            sa.sql.func.trim(ns) == ns, ns != u''),),
        # label is not empty and is not surrounded by space characters
        sa.CheckConstraint(sa.sql.and_(
            sa.sql.func.trim(label) == label, label != u''),),)

    def __unicode__(self):
        return self.label

    def __lt__(self, other):
        return unicode(self).lower().__lt__(unicode(other).lower())

    def __repr__(self):
        cls = self.__class__
        return '<{mod}.{cls} id={t.id!r} ns={t.ns!r} label={t.label!r} at 0x{addr:x}>'.format(
            mod=cls.__module__,
            cls=cls.__name__,
            t=self,
            addr=id(self),)



# -*- coding: utf-8 -*-
"""
View decorators for controlling some aspect of HTTP, mainly: Cache headers
"""
from __future__ import absolute_import, division, print_function

from functools import wraps

from flask import make_response


def nocache(view):

    @wraps(view)
    def _nocache(*args, **kwargs):
        resp = make_response(view(*args, **kwargs))
        resp.headers['Cache-Control'] = 'no-cache'
        return resp

    return _nocache



""" Application settings service
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from .service import SettingsService
from .models import register

__all__ = ['SettingsService', 'register']



"""Social graph API.

May or may not be needed, since the followers / followees notions are already
implemented in the core model.
"""
from __future__ import absolute_import, division, print_function


class FollowService(object):

    def get_followers(self, user):
        pass

    def get_followees(self, user):
        pass



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

from flask import current_app, jsonify, send_file
from flask_login import current_user
from flask_wtf.file import FileField, file_required
from werkzeug.exceptions import BadRequest, NotFound
from werkzeug.utils import secure_filename

from abilian.web import csrf, url_for
from abilian.web.blueprints import Blueprint
from abilian.web.forms import Form
from abilian.web.views import JSONView, View

bp = Blueprint('uploads', __name__, url_prefix='/upload')


class UploadForm(Form):

    file = FileField(validators=(file_required(),))


class BaseUploadsView(object):

    def prepare_args(self, args, kwargs):
        args, kwargs = super(BaseUploadsView, self).prepare_args(args, kwargs)
        self.uploads = current_app.extensions['uploads']
        self.user = current_user._get_current_object()
        return args, kwargs


class NewUploadView(BaseUploadsView, JSONView):
    """
    Upload a new file
    """
    methods = ['POST', 'PUT']
    decorators = (csrf.support_graceful_failure,)

    #: file handle to be returned
    handle = None

    def data(self, *args, **kwargs):
        return {
            'handle': self.handle,
            'url': url_for('.handle', handle=self.handle),
        }

    def post(self, *args, **kwargs):
        form = UploadForm()

        if not form.validate():
            raise BadRequest('File is missing.')

        uploaded = form['file'].data
        filename = secure_filename(uploaded.filename)
        mimetype = uploaded.mimetype
        self.handle = self.uploads.add_file(self.user,
                                            uploaded,
                                            filename=filename,
                                            mimetype=mimetype)
        return self.get(*args, **kwargs)

    def put(self, *args, **kwargs):
        return self.post(*args, **kwargs)


bp.add_url_rule('/', view_func=NewUploadView.as_view('new_file',),)


class UploadView(BaseUploadsView, View):
    """
    Manage an uploaded file: download, delete
    """
    methods = ['GET', 'DELETE']
    decorators = (csrf.support_graceful_failure,)

    def get(self, handle, *args, **kwargs):
        file_obj = self.uploads.get_file(self.user, handle)

        if file_obj is None:
            raise NotFound()

        metadata = self.uploads.get_metadata(self.user, handle)
        filename = metadata.get('filename', handle)
        content_type = metadata.get('mimetype')
        stream = file_obj.open('rb')

        return send_file(stream,
                         as_attachment=True,
                         attachment_filename=filename,
                         mimetype=content_type,
                         cache_timeout=0,
                         add_etags=False,
                         conditional=False)

    def delete(self, handle, *args, **kwargs):
        if self.uploads.get_file(self.user, handle) is None:
            raise NotFound()

        self.uploads.remove_file(self.user, handle)
        return jsonify({'success': True})


bp.add_url_rule('/<string:handle>', view_func=UploadView.as_view('handle',),)



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

from flask import g

from abilian.testing import BaseTestCase


class TestIdGenerator(BaseTestCase):

    def test_id_generator(self):
        # inside app context: id_generator installed
        assert hasattr(g, 'id_generator')

        # test sucessives values
        assert next(g.id_generator) == 1
        assert next(g.id_generator) == 2
        assert next(g.id_generator) == 3

        gen = g.id_generator
        with self.app.app_context():
            # new app context: new id generator.
            # Note: if this behaviour is not desired and got changed, this test will
            # flag behaviour change.
            assert g.id_generator is not gen
            assert next(g.id_generator) == 1

        # app context popped: first id generator 'restored'
        assert g.id_generator is gen
        assert next(g.id_generator) == 4



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import sqlalchemy as sa
from wtforms.fields import BooleanField, StringField
from wtforms.validators import ValidationError

from abilian.core.models.subjects import Group
from abilian.i18n import _, _l
from abilian.services.security.models import Role
from abilian.web.forms import Form, widgets
from abilian.web.forms.fields import QuerySelect2Field, Select2MultipleField
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import optional, required


class BaseUserAdminForm(Form):

    email = StringField(
        _l(u'Email'),
        description=_l(u'Users log in with their email address.'),
        view_widget=widgets.EmailWidget(),
        filters=(strip,),
        validators=[required()])
    first_name = StringField(
        _l(u'First Name'),
        description=_l(u'ex: John'),
        filters=(strip,),
        validators=[required()])
    last_name = StringField(
        _l(u'Last Name'),
        description=_l(u'ex: Smith'),
        filters=(strip,),
        validators=[required()])

    can_login = BooleanField(
        _l(u'Login enabled'),
        description=_l(u'If unchecked, user will not be able to connect.'),
        widget=widgets.BooleanWidget())

    groups = QuerySelect2Field(
        _l(u'Groups'),
        validators=(optional(),),
        multiple=True,
        collection_class=set,
        query_factory=
        lambda: Group.query.order_by(sa.sql.func.lower(Group.name).asc()),
        get_label='name',)

    roles = Select2MultipleField(
        _l(u'Roles'),
        description=_l(
            u'Prefer groups to manage access rights. Directly assigning roles '
            u'to users is possible but discouraged.'),
        choices=lambda: [(r.name, r.label) for r in Role.assignable_roles()],)

    password = StringField(
        _l(u'New Password'),
        description=_l(u'If empty the current password will not be changed.'),
        widget=widgets.PasswordInput(autocomplete='off'))


class UserAdminForm(BaseUserAdminForm):

    confirm_password = StringField(
        _l(u'Confirm new password'),
        widget=widgets.PasswordInput(autocomplete='off'))

    def validate_password(self, field):
        pwd = field.data
        confirmed = self['confirm_password'].data

        if pwd != confirmed:
            raise ValidationError(
                _(u'Passwords differ. Ensure you have typed same password in both'
                  u' "password" field and "confirm password" field.'))


class UserCreateForm(BaseUserAdminForm):

    password = StringField(
        _l(u'Password'),
        description=_l(u'If empty a random password will be generated.'),
        widget=widgets.PasswordInput(autocomplete='off'))



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import abc

import sqlalchemy as sa
from sqlalchemy import Column, ForeignKey, Integer, UnicodeText
from sqlalchemy.orm import backref, relationship

from abilian.core.entities import Entity

from .blob import Blob

#: name of backref on target :class:`Entity` object
ATTRIBUTE = '__attachments__'


class SupportAttachment(object):
    __metaclass__ = abc.ABCMeta


def register(cls):
    """Register an :class:`~.Entity` as a attachmentable class.

    Can be used as a class decorator:

    .. code-block:: python

      @attachment.register
      class MyContent(Entity):
          ....
    """
    if not issubclass(cls, Entity):
        raise ValueError(
            'Class must be a subclass of abilian.core.entities.Entity')

    SupportAttachment.register(cls)
    return cls


def is_support_attachments(obj):
    """
    :param obj: a class or instance
    """
    if isinstance(obj, type):
        return issubclass(obj, SupportAttachment)

    if not isinstance(obj, SupportAttachment):
        return False

    if obj.id is None:
        return False

    return True


def for_entity(obj, check_support_attachments=False):
    """Return attachments on an entity.
    """
    if check_support_attachments and not is_support_attachments(obj):
        return []

    return getattr(obj, ATTRIBUTE)


class Attachment(Entity):
    """An Attachment owned by an :class:`Entity`.
    """
    __auditable_entity__ = ('entity', 'attachment', ('id', 'name'))

    @sa.ext.declarative.declared_attr
    def __mapper_args__(cls):
        # we cannot use super(Attachment, cls): declared_attr happens during class
        # construction. super(cls, cls) could work; as long as `cls` is not a
        # subclass of `Attachment`: it would enter into an infinite loop.
        #
        # Entity.__mapper_args__ calls the descriptor with 'Entity', not `cls`.
        args = Entity.__dict__['__mapper_args__'].fget(cls)
        args['order_by'] = cls.created_at
        return args

    entity_id = Column(Integer, ForeignKey(Entity.id), nullable=False)

    #: owning entity
    entity = relationship(Entity,
                          lazy='immediate',
                          foreign_keys=[entity_id],
                          backref=backref(ATTRIBUTE,
                                          lazy='select',
                                          order_by='Attachment.created_at',
                                          cascade="all, delete-orphan",))

    blob_id = Column(Integer, sa.ForeignKey(Blob.id), nullable=False)
    #: file. Stored in a :class:`Blob`
    blob = relationship(Blob, cascade='all, delete', foreign_keys=[blob_id])

    description = Column(UnicodeText(),
                         nullable=False,
                         default=u'',
                         server_default=u'',)

    def __repr__(self):
        class_ = self.__class__
        mod_ = class_.__module__
        classname = class_.__name__
        return '<{}.{} instance at 0x{:x} entity id={!r}'\
            .format(mod_, classname, id(self), self.entity_id)


@sa.event.listens_for(Attachment, 'before_insert', propagate=True)
@sa.event.listens_for(Attachment, 'before_update', propagate=True)
def set_attachment_name(mapper, connection, target):
    if target.name:
        return

    blob = target.blob
    if not blob:
        return

    filename = blob.meta.get('filename')
    if filename is not None:
        target.name = filename



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, division

from .extension import TagsExtension
from .criterion import TagCriterion

__all__ = ['TagsExtension', 'TagCriterion']


def register_plugin(app):
    TagsExtension(app=app)
